/*
 * Copyright (C) 2020 Hermann Meyer
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 * --------------------------------------------------------------------------
 */


#include <cstdlib>
#include <cmath>
#include <iostream>
#include <cstring>
#include <unistd.h>

///////////////////////// DENORMAL PROTECTION WITH SSE /////////////////

#ifdef NOSSE
#undef __SSE__
#endif

#ifdef __SSE__
#include <immintrin.h>
#ifndef _IMMINTRIN_H_INCLUDED
#include <fxsrintrin.h>
#endif
/* On Intel set FZ (Flush to Zero) and DAZ (Denormals Are Zero)
   flags to avoid costly denormals */
#ifdef __SSE3__
#ifndef _PMMINTRIN_H_INCLUDED
#include <pmmintrin.h>
#endif
#else
#ifndef _XMMINTRIN_H_INCLUDED
#include <xmmintrin.h>
#endif
#endif //__SSE3__

#endif //__SSE__

///////////////////////// MACRO SUPPORT ////////////////////////////////

#define __rt_func __attribute__((section(".rt.text")))
#define __rt_data __attribute__((section(".rt.data")))

///////////////////////// FAUST SUPPORT ////////////////////////////////

#define FAUSTFLOAT float
#ifndef N_
#define N_(String) (String)
#endif
#define max(x, y) (((x) > (y)) ? (x) : (y))
#define min(x, y) (((x) < (y)) ? (x) : (y))

#define always_inline inline __attribute__((always_inline))

#ifndef signbit
#define signbit(x) std::signbit(x)
#endif

template<class T> inline T mydsp_faustpower2_f(T x) {return (x * x);}
template<class T> inline T mydsp_faustpower3_f(T x) {return ((x * x) * x);}
template<class T> inline T mydsp_faustpower4_f(T x) {return (((x * x) * x) * x);}
template<class T> inline T mydsp_faustpower5_f(T x) {return ((((x * x) * x) * x) * x);}
template<class T> inline T mydsp_faustpower6_f(T x) {return (((((x * x) * x) * x) * x) * x);}

////////////////////////////// LOCAL INCLUDES //////////////////////////

#include "gx_melmatcheq.h"        // define struct PortIndex
#include "gx_pluginlv2.h"   // define struct PluginLV2
#include "MelEQ.cc"    // dsp class generated by faust -> dsp2cc

////////////////////////////// PLUG-IN CLASS ///////////////////////////

namespace matcheq {

class DenormalProtection
{
private:
#ifdef __SSE__
  uint32_t  mxcsr_mask;
  uint32_t  mxcsr;
  uint32_t  old_mxcsr;
#endif

public:
  inline void set_() {
#ifdef __SSE__
    old_mxcsr = _mm_getcsr();
    mxcsr = old_mxcsr;
    _mm_setcsr((mxcsr | _MM_DENORMALS_ZERO_MASK | _MM_FLUSH_ZERO_MASK) & mxcsr_mask);
#endif
  };
  inline void reset_() {
#ifdef __SSE__
    _mm_setcsr(old_mxcsr);
#endif
  };

  inline DenormalProtection()
  {
#ifdef __SSE__
    mxcsr_mask = 0xffbf; // Default MXCSR mask
    mxcsr      = 0;
    uint8_t fxsave[512] __attribute__ ((aligned (16))); // Structure for storing FPU state with FXSAVE command

    memset(fxsave, 0, sizeof(fxsave));
    __builtin_ia32_fxsave(&fxsave);
    uint32_t mask = *(reinterpret_cast<uint32_t *>(&fxsave[0x1c])); // Obtain the MXCSR mask from FXSAVE structure
    if (mask != 0)
        mxcsr_mask = mask;
#endif
  };

  inline ~DenormalProtection() {};
};

class Gx_matcheq_
{
private:
  // pointer to buffer
  float*          output;
  float*          input;
  // pointer to dsp class
  PluginLV2*      matcheq;
  DenormalProtection MXCSR;

  // bypass ramping
  float*          bypass;
  uint32_t        bypass_;
  float*          match2;
  uint32_t        match2_;
  float*          clear;
  uint32_t        clear_;

  bool            needs_ramp_down;
  bool            needs_ramp_up;
  float           ramp_down;
  float           ramp_up;
  float           ramp_up_step;
  float           ramp_down_step;
  bool            bypassed;
  bool            no_clear;

  // private functions
  inline void run_dsp_(uint32_t n_samples);
  inline void connect_(uint32_t port,void* data);
  inline void init_dsp_(uint32_t rate);
  inline void connect_all__ports(uint32_t port, void* data);
  inline void activate_f();
  inline void clean_up();
  inline void deactivate_f();

public:
  // LV2 Descriptor
  static const LV2_Descriptor descriptor;
  // static wrapper to private functions
  static void deactivate(LV2_Handle instance);
  static void cleanup(LV2_Handle instance);
  static void run(LV2_Handle instance, uint32_t n_samples);
  static void activate(LV2_Handle instance);
  static void connect_port(LV2_Handle instance, uint32_t port, void* data);
  static LV2_Handle instantiate(const LV2_Descriptor* descriptor,
                                double rate, const char* bundle_path,
                                const LV2_Feature* const* features);
  Gx_matcheq_();
  ~Gx_matcheq_();
};

// constructor
Gx_matcheq_::Gx_matcheq_() :
  output(NULL),
  input(NULL),
  matcheq(matcheq::plugin()),
  MXCSR(),
  bypass(0),
  bypass_(2),
  match2(0),
  match2_(0),
  clear(0),
  clear_(0),
  needs_ramp_down(false),
  needs_ramp_up(false),
  bypassed(false),
  no_clear(true) {};

// destructor
Gx_matcheq_::~Gx_matcheq_()
{
  // just to be sure the plug have given free the allocated mem
  // it didn't hurd if the mem is already given free by clean_up()
  if (matcheq->activate_plugin !=0)
    matcheq->activate_plugin(false, matcheq);
  // delete DSP class
  matcheq->delete_instance(matcheq);
};

///////////////////////// PRIVATE CLASS  FUNCTIONS /////////////////////

void Gx_matcheq_::init_dsp_(uint32_t rate)
{
  // set values for internal ramping
  ramp_down_step = 32 * (256 * rate) / 48000; 
  ramp_up_step = ramp_down_step;
  ramp_down = ramp_down_step;
  ramp_up = 0.0;

  matcheq->set_samplerate(rate, matcheq); // init the DSP class
}

// connect the Ports used by the plug-in class
void Gx_matcheq_::connect_(uint32_t port,void* data)
{
  switch ((PortIndex)port)
    {
    case EFFECTS_OUTPUT:
      output = static_cast<float*>(data);
      break;
    case EFFECTS_INPUT:
      input = static_cast<float*>(data);
      break;
    case BYPASS: 
      bypass = static_cast<float*>(data); // , 0.0, 0.0, 1.0, 1.0 
      break;
    case MATCH2: 
      match2 = static_cast<float*>(data); // , 0.0, 0.0, 1.0, 1.0 
      break;
    case CLEAR: 
      clear = static_cast<float*>(data); // , 0.0, 0.0, 1.0, 1.0 
      break;
    default:
      break;
    }
}

void Gx_matcheq_::activate_f()
{
  // allocate the internal DSP mem
  if (matcheq->activate_plugin !=0)
    matcheq->activate_plugin(true, matcheq);
}

void Gx_matcheq_::clean_up()
{
  // delete the internal DSP mem
  if (matcheq->activate_plugin !=0)
    matcheq->activate_plugin(false, matcheq);
}

void Gx_matcheq_::deactivate_f()
{
  // delete the internal DSP mem
  if (matcheq->activate_plugin !=0)
    matcheq->activate_plugin(false, matcheq);
}

void Gx_matcheq_::run_dsp_(uint32_t n_samples)
{
  MXCSR.set_();

  FAUSTFLOAT buf[n_samples];
  // do inplace processing at default
  if (output != input)
    memcpy(output, input, n_samples*sizeof(float));
  // check if bypass is pressed
  if (bypass_ != static_cast<uint32_t>(*(bypass))) {
    bypass_ = static_cast<uint32_t>(*(bypass));
    if (!bypass_) {
      needs_ramp_down = true;
      needs_ramp_up = false;
      no_clear = false;
    } else {
      needs_ramp_down = false;
      needs_ramp_up = true;
      bypassed = false;
    }
  }
  if (match2_ != static_cast<uint32_t>(*(match2))) {
    match2_ = static_cast<uint32_t>(*(match2));
   // if (!match2_) {
      needs_ramp_down = true;
      needs_ramp_up = true;
      no_clear = true;
    //}
  }
  if (clear_ != static_cast<uint32_t>(*(clear))) {
    clear_ = static_cast<uint32_t>(*(clear));
    if (!clear_) {
      needs_ramp_down = true;
      needs_ramp_up = true;
      no_clear = true;
    }
  }

  if (needs_ramp_down || needs_ramp_up) {
       memcpy(buf, input, n_samples*sizeof(float));
  }
  
  if (!bypassed) {
      matcheq->mono_audio(static_cast<int>(n_samples), output, output, matcheq);
  }

  // check if ramping is needed
  if (needs_ramp_down) {
    float fade = 0;
    for (uint32_t i=0; i<n_samples; i++) {
      if (ramp_down >= 0.0) {
        --ramp_down; 
      }
      fade = max(0.0,ramp_down) /ramp_down_step ;
      output[i] = output[i] * fade + buf[i] * (1.0 - fade);
    }

    if (ramp_down <= 0.0) {
      // when ramped down, clear buffer from matcheq class
      if (!no_clear) {
        matcheq->clear_state(matcheq);
        bypassed = true;
      }
      needs_ramp_down = false;
      ramp_down = ramp_down_step;
      ramp_up = 0.0;
    } else {
      ramp_up = ramp_down;
    }

  } else if (needs_ramp_up) {
    float fade = 0;
    for (uint32_t i=0; i<n_samples; i++) {
      if (ramp_up < ramp_up_step) {
        ++ramp_up ;
      }
      fade = min(ramp_up_step,ramp_up) /ramp_up_step ;
      output[i] = output[i] * fade + buf[i] * (1.0 - fade);
    }

    if (ramp_up >= ramp_up_step) {
      needs_ramp_up = false;
      ramp_up = 0.0;
      ramp_down = ramp_down_step;
    } else {
      ramp_down = ramp_up;
    }
  }

  MXCSR.reset_();
}

void Gx_matcheq_::connect_all__ports(uint32_t port, void* data)
{
  // connect the Ports used by the plug-in class
  connect_(port,data); 
  // connect the Ports used by the DSP class
  matcheq->connect_ports(port,  data, matcheq);
}

////////////////////// STATIC CLASS  FUNCTIONS  ////////////////////////

LV2_Handle 
Gx_matcheq_::instantiate(const LV2_Descriptor* descriptor,
                            double rate, const char* bundle_path,
                            const LV2_Feature* const* features)
{
  // init the plug-in class
  Gx_matcheq_ *self = new Gx_matcheq_();
  if (!self) {
    return NULL;
  }

  self->init_dsp_((uint32_t)rate);

  return (LV2_Handle)self;
}

void Gx_matcheq_::connect_port(LV2_Handle instance, 
                                   uint32_t port, void* data)
{
  // connect all ports
  static_cast<Gx_matcheq_*>(instance)->connect_all__ports(port, data);
}

void Gx_matcheq_::activate(LV2_Handle instance)
{
  // allocate needed mem
  static_cast<Gx_matcheq_*>(instance)->activate_f();
}

void Gx_matcheq_::run(LV2_Handle instance, uint32_t n_samples)
{
  // run dsp
  static_cast<Gx_matcheq_*>(instance)->run_dsp_(n_samples);
}

void Gx_matcheq_::deactivate(LV2_Handle instance)
{
  // free allocated mem
  static_cast<Gx_matcheq_*>(instance)->deactivate_f();
}

void Gx_matcheq_::cleanup(LV2_Handle instance)
{
  // well, clean up after us
  Gx_matcheq_* self = static_cast<Gx_matcheq_*>(instance);
  self->clean_up();
  delete self;
}

const LV2_Descriptor Gx_matcheq_::descriptor =
{
  GXPLUGIN_URI "#_melmatcheq_",
  Gx_matcheq_::instantiate,
  Gx_matcheq_::connect_port,
  Gx_matcheq_::activate,
  Gx_matcheq_::run,
  Gx_matcheq_::deactivate,
  Gx_matcheq_::cleanup,
  NULL
};


} // end namespace matcheq

////////////////////////// LV2 SYMBOL EXPORT ///////////////////////////

extern "C"
LV2_SYMBOL_EXPORT
const LV2_Descriptor*
lv2_descriptor(uint32_t index)
{
  switch (index)
    {
    case 0:
      return &matcheq::Gx_matcheq_::descriptor;
    default:
      return NULL;
    }
}

///////////////////////////// FIN //////////////////////////////////////
